// MemoryPool.cpp: implementation of the CMemoryPool class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MemoryPool.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////

CMemoryPool::CMemoryPool()
{
	m_pMemCurrent = NULL;
	m_dwMemCurrentSize = 0;

	MEMORYSTATUS ms;
	GlobalMemoryStatus(&ms);
	if (ms.dwAvailPhys < 64 * 1024 * 1024)
		m_dwMemSizeMax = ms.dwAvailPhys / 2;
	else
		m_dwMemSizeMax = ms.dwAvailPhys - 32 * 1024 * 1024;
}

CMemoryPool::~CMemoryPool()
{
	if (m_pMemCurrent != NULL)
		delete[] m_pMemCurrent;
}

void* CMemoryPool::AllocateMemoryChunk(DWORD dwFileLength, DWORD& dwAllocatedSize)
{
	// Calculate dwTargetSize
	DWORD dwTargetSize;
	if (dwFileLength > m_dwMemSizeMax)
		dwTargetSize = m_dwMemSizeMax;
	else
		dwTargetSize = dwFileLength;

	// Determine what to do
	if (dwTargetSize <= m_dwMemCurrentSize)
	{
		//printf("debug: using existing memory chunk\n");
		dwAllocatedSize = dwTargetSize;
		return m_pMemCurrent;
	}
	else
	{
		// Free existing memory
		if (m_pMemCurrent != NULL)
			delete[] m_pMemCurrent;
		m_dwMemCurrentSize = 0;

		// Allocate new memory chunk
		//printf("debug: allocating new memory chunk\n");
		void* pMem = new BYTE[dwTargetSize];
		if (pMem != NULL)
		{
			m_pMemCurrent = pMem;
			m_dwMemCurrentSize = dwTargetSize;

			dwAllocatedSize = m_dwMemCurrentSize;
			return m_pMemCurrent;
		}
		else
		{
			dwAllocatedSize = 0;
			return NULL;
		}
	}
}
