// RainbowCrack.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "RainbowCrack.h"
#include "../RainbowPublic/TradeoffContext.h"
#include "HashContainer.h"
#include "RainbowChainWalkContainer.h"
#include "MemoryPool.h"
#include "../RainbowPublic/Public.h"

struct RainbowChain
{
	unsigned __int64 nIndexS;
	unsigned __int64 nIndexE;
};

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////

float g_fTotalDiskAccessTime = 0.0f;
float g_fTotalCryptanalysisTime = 0.0f;
int g_nTotalChainWalkStep = 0;
int g_nTotalFalseAlarm = 0;
int g_nTotalFalseAlarmStep = 0;

/////////////////////////////////////////////////////////////////////////////

void GetFileList(CString sWildcharPathName, CStringArray& arrPathName)
{
	arrPathName.RemoveAll();

	CFileFind finder;
	BOOL bWorking = finder.FindFile(sWildcharPathName);
	while (bWorking)
	{
		bWorking = finder.FindNextFile();
		if (finder.IsDots())
			continue;
		if (finder.IsDirectory())
			continue;
		arrPathName.Add(finder.GetFilePath());
	}
}

BOOL ParseFileName(CString sFileName, CString& sPlainCharset, int& nRainbowTableIndex, int& nRainbowChainLength, int& nRainbowChainCount)
{
	if (sFileName.Right(3) != ".rt")
		return FALSE;

	int n1 = sFileName.Find("_");
	int n2 = sFileName.Find("_", n1 + 1);
	int n3 = sFileName.Find("_", n2 + 1);
	int n4 = sFileName.Find("x", n3 + 1);
	int n5 = sFileName.Find("_", n4 + 1);
	if (n1 == -1 || n2 == -1 || n3 == -1 || n4 == -1 || n5 == -1)
		return FALSE;
	CString sCipherType = sFileName.Left(n1);
	if (sCipherType != "lm")
		return FALSE;

	sPlainCharset               = sFileName.Mid(n1 + 1, n2 - n1 - 1);
	CString sRainbowTableIndex  = sFileName.Mid(n2 + 1, n3 - n2 - 1);
	CString sRainbowChainLength = sFileName.Mid(n3 + 1, n4 - n3 - 1);
	CString sRainbowChainCount  = sFileName.Mid(n4 + 1, n5 - n4 - 1);
	nRainbowTableIndex  = atoi(sRainbowTableIndex);
	nRainbowChainLength = atoi(sRainbowChainLength);
	nRainbowChainCount  = atoi(sRainbowChainCount);

	return TRUE;
}

void ParseHash(CString sHash, BYTE Hash[8])
{
	int i;
	for (i = 0; i < 8; i++)
	{
		char str[3] = {0};
		str[0] = sHash[i * 2];
		str[1] = sHash[i * 2 + 1];
		int value;
		sscanf(str, "%02x", &value);
		Hash[i] = BYTE(value);
	}
}

BOOL LoadPwdDumpFile(CString sPathName, CStringArray& arrUserName, CStringArray& arrLMHash)
{
	arrUserName.RemoveAll();
	arrLMHash.RemoveAll();

	CStdioFile file;
	if (file.Open(sPathName, CFile::modeRead | CFile::typeText))
	{
		CString sLine;
		while (file.ReadString(sLine))
		{
			sLine.TrimLeft();
			sLine.TrimRight();
			int n1 = sLine.Find(":");
			int n2 = sLine.Find(":", n1 + 1);
			int n3 = sLine.Find(":", n2 + 1);
			if (n1 != -1 && n2 != -1 && n3 != -1)
			{
				CString sUserName = sLine.Left(n1);
				CString sLMHash   = sLine.Mid(n2 + 1, n3 - n2 - 1);
				sLMHash.MakeLower();
				if (sLMHash.GetLength() == 32 && sLMHash != "no password*********************")
				{
					arrUserName.Add(sUserName);
					arrLMHash.Add(sLMHash);
				}
			}
		}
		file.Close();
	}
	else
		return FALSE;

	return TRUE;
}

int BinarySearch(RainbowChain* pChain, int nRainbowChainCount, unsigned __int64 nIndex)
{
	int nLow = 0;
	int nHigh = nRainbowChainCount - 1;
	while (nLow <= nHigh)
	{
		int nMid = (nLow + nHigh) / 2;
		if (nIndex == pChain[nMid].nIndexE)
			return nMid;
		else if (nIndex < pChain[nMid].nIndexE)
			nHigh = nMid - 1;
		else
			nLow = nMid + 1;
	}

	return -1;
}

void GetChainIndexRangeWithSameEndpoint(RainbowChain* pChain,
										int nRainbowChainCount,
										int nChainIndex,
										int& nChainIndexFrom,
										int& nChainIndexTo)
{
	nChainIndexFrom = nChainIndex;
	nChainIndexTo   = nChainIndex;
	while (nChainIndexFrom > 0)
	{
		if (pChain[nChainIndexFrom - 1].nIndexE == pChain[nChainIndex].nIndexE)
			nChainIndexFrom--;
		else
			break;
	}
	while (nChainIndexTo < nRainbowChainCount - 1)
	{
		if (pChain[nChainIndexTo + 1].nIndexE == pChain[nChainIndex].nIndexE)
			nChainIndexTo++;
		else
			break;
	}
}

BOOL CheckAlarm(RainbowChain* pChain, int nChainIndex, int nChainPos, BYTE TargetHash[8], CString& sResult)
{
	CTradeoffContext tc;
	tc.SetIndex(pChain[nChainIndex].nIndexS);
	int i;
	for (i = 0; i < nChainPos; i++)
	{
		tc.IndexToPlain();
		tc.PlainToHash();
		tc.HashToIndex(i);
	}
	tc.IndexToPlain();
	tc.PlainToHash();
	if (tc.CheckHash(TargetHash) == TRUE)
	{
		sResult = tc.GetPlain();
		return TRUE;
	}

	return FALSE;
}

void RainbowCrack(RainbowChain* pChain, int nRainbowChainLength, int nRainbowChainCount,
				  CHashContainer& hc, CRainbowChainWalkContainer& rcwc)
{
	CStringArray arrLeftHash;
	hc.GetAllLeftHash(arrLeftHash);
	if (arrLeftHash.GetSize() == 1)
		printf("searching for %d hash ...\n",   arrLeftHash.GetSize());
	else
		printf("searching for %d hashes ...\n", arrLeftHash.GetSize());

	// Search the hashes one by one
	int nChainWalkStep = 0;
	int nFalseAlarm = 0;
	int nFalseAlarmStep = 0;
	int nHashIndex;
	for (nHashIndex = 0; nHashIndex < arrLeftHash.GetSize(); nHashIndex++)
	{
		BYTE TargetHash[8];
		ParseHash(arrLeftHash[nHashIndex], TargetHash);

		// Request RainbowChainWalk
		RainbowChainWalk rcw;
		memcpy(rcw.Hash, TargetHash, 8);
		BOOL fNewlyGeneratedWalk = rcwc.RequestWalk(CTradeoffContext::GetPlainCharset(),
													CTradeoffContext::GetRainbowTableIndex(),
													nRainbowChainLength,
													rcw);

		//if (fNewlyGeneratedWalk == TRUE)
		//	printf("debug: using newly generated walk for %s\n", arrLeftHash[nHashIndex]);
		//else
		//	printf("debug: using existing walk for %s\n", arrLeftHash[nHashIndex]);

		// Walk the chain
		int nPos;
		for (nPos = nRainbowChainLength - 2; nPos >= 0; nPos--)		// Which traverse is faster?
		//for (nPos = 0; nPos <= nRainbowChainLength - 2; nPos++)	// Which traverse is faster?
		{
			// Use newly generated version or saved version
			unsigned __int64 nIndexEOfCurrentPos;
			if (fNewlyGeneratedWalk == TRUE)
			{
				CTradeoffContext tc;
				tc.SetHash(TargetHash);
				tc.HashToIndex(nPos);
				int i;
				for (i = nPos + 1; i <= nRainbowChainLength - 2; i++)
				{
					tc.IndexToPlain();
					tc.PlainToHash();
					tc.HashToIndex(i);
				}
				rcw.pIndexE[nPos]   = tc.GetIndex();
				nIndexEOfCurrentPos = tc.GetIndex();
				nChainWalkStep += nRainbowChainLength - 2 - nPos;
			}
			else
				nIndexEOfCurrentPos = rcw.pIndexE[nPos];

			// Search matching nIndexE
			int nChainIndex = BinarySearch(pChain, nRainbowChainCount, nIndexEOfCurrentPos);
			if (nChainIndex != -1)
			{
				int nChainIndexFrom, nChainIndexTo;
				GetChainIndexRangeWithSameEndpoint(pChain, nRainbowChainCount, nChainIndex, nChainIndexFrom, nChainIndexTo);

				int i;
				for (i = nChainIndexFrom; i <= nChainIndexTo; i++)
				{
					CString sResult;
					if (CheckAlarm(pChain, i, nPos, TargetHash, sResult) == TRUE)
					{
						printf("plaintext of %s is %s\n", arrLeftHash[nHashIndex], sResult);
						hc.SetPlainOfHash(arrLeftHash[nHashIndex], sResult);
						//printf("debug: discarding walk for %s\n", arrLeftHash[nHashIndex]);
						rcwc.DiscardWalk(rcw);
						goto NEXT_HASH;
					}
					else
					{
						nFalseAlarm++;
						nFalseAlarmStep += nPos + 1;
					}
				}
			}
		}
NEXT_HASH:;
	}
	//printf("debug: chain walk step: %d\n",	nChainWalkStep);
	//printf("debug: false alarm: %d\n",		nFalseAlarm);
	//printf("debug: false alarm step: %d\n",	nFalseAlarmStep);

	g_nTotalChainWalkStep += nChainWalkStep;
	g_nTotalFalseAlarm += nFalseAlarm;
	g_nTotalFalseAlarmStep += nFalseAlarmStep;
}

void SearchRainbowTable(CString sPathName, CHashContainer& hc, CRainbowChainWalkContainer& rcwc)
{
	// Parse filename
	int n = sPathName.ReverseFind('\\');
	CString sFileName = sPathName.Mid(n + 1);
	CString sPlainCharset;
	int nRainbowTableIndex;
	int nRainbowChainLength;
	int nRainbowChainCount;
	if (ParseFileName(sFileName, sPlainCharset, nRainbowTableIndex, nRainbowChainLength, nRainbowChainCount) == FALSE)
	{
		printf("filename %s not identified, skipping ...\n", sFileName);
		return;
	}
	if (CTradeoffContext::SetPlainCharset(sPlainCharset) == FALSE)
	{
		printf("charset %s not supported, skipping ...\n", sPlainCharset);
		return;
	}
	if (CTradeoffContext::SetRainbowTableIndex(nRainbowTableIndex) == FALSE)
	{
		printf("invalid rainbow table index, skipping ...\n");
		return;
	}

	// Open file
	CFile file;
	if (file.Open(sPathName, CFile::modeRead) == 0)
	{
		printf("can't open %s, skipping ...\n", sFileName);
		return;
	}

	// Search key
	if (file.GetLength() % 16 != 0 || nRainbowChainCount != int(file.GetLength() / 16))
		printf("file length of %s does not match with rainbow chain length, skipping ...\n", sFileName);
	else
	{
		static CMemoryPool mp;
		DWORD dwAllocatedSize;
		RainbowChain* pChain = (RainbowChain*)mp.AllocateMemoryChunk(file.GetLength(), dwAllocatedSize);
		dwAllocatedSize -= dwAllocatedSize % 16;	// Round to 16-byte boundary
		if (pChain != NULL)
		{
			BOOL fVerified = FALSE;
			while (TRUE)
			{
				if (file.GetPosition() == file.GetLength())
					break;

				// Load
				printf("reading %s ...\n", sFileName);
				DWORD dwTick1 = GetTickCount();
				DWORD dwDataRead = file.Read(pChain, dwAllocatedSize);
				DWORD dwTick2 = GetTickCount();
				float fDiskAccessTime = (dwTick2 - dwTick1) * 1.0f / 1000;
				printf("%u bytes read, disk access time: %.2f s\n", dwDataRead, fDiskAccessTime);
				g_fTotalDiskAccessTime += fDiskAccessTime;

				int nRainbowChainCountRead = dwDataRead / 16;

				// Verify file - one time is enough
				if (fVerified == FALSE)
				{
					printf("verifying the file ...\n");

					int nIndexToVerify = nRainbowChainCountRead / 2;
					CTradeoffContext tc;
					tc.SetIndex(pChain[nIndexToVerify].nIndexS);
					int nPos;
					for (nPos = 0; nPos < nRainbowChainLength - 1; nPos++)
					{
						tc.IndexToPlain();
						tc.PlainToHash();
						tc.HashToIndex(nPos);
					}
					if (tc.GetIndex() != pChain[nIndexToVerify].nIndexE)
					{
						printf("rainbow chain length verify fail, skipping ...\n");
						break;
					}

					int i;
					for (i = 0; i < nRainbowChainCountRead - 1; i++)
					{
						if (pChain[i].nIndexE > pChain[i + 1].nIndexE)
							break;
					}
					if (i != nRainbowChainCountRead - 1)
					{
						printf("%s is not sorted, skipping ...\n", sFileName);
						break;
					}

					fVerified = TRUE;
				}

				// Search key
				dwTick1 = GetTickCount();
				RainbowCrack(pChain, nRainbowChainLength, nRainbowChainCountRead, hc, rcwc);
				dwTick2 = GetTickCount();
				float fCryptanalysisTime = (dwTick2 - dwTick1) * 1.0f / 1000;
				printf("cryptanalysis time: %.2f s\n", fCryptanalysisTime);
				g_fTotalCryptanalysisTime += fCryptanalysisTime;

				// Any hash left?
				if (hc.AnyHashLeft() == FALSE)
					break;
			}
		}
		else
			printf("memory allocation fail when loading %s, skipping ...\n", sFileName);
	}

	// Close file
	file.Close();
}

void Usage()
{
	Logo();

	printf("usage: rcrack rainbow_table_filename -h lmhash\n");
	printf("       rcrack rainbow_table_filename -f pwdump_file\n");
	printf("rainbow_table_filename: filename of the rainbow table, wildchar(*, ?) supported\n");
	printf("-h lmhash:              use raw lanmanager hash as input\n");
	printf("-f pwdump_file:         use pwdump file as input\n");
}

CWinApp theApp;
int _tmain(int argc, TCHAR* argv[], TCHAR* envp[])
{
	if (!AfxWinInit(::GetModuleHandle(NULL), NULL, ::GetCommandLine(), 0))
		return 0;

	if (argc != 4)
	{
		Usage();
		return 0;
	}
	CString sWildcharPathName = argv[1];
	CString sInputType = argv[2];
	CString sInput     = argv[3];

	// Load UserName, LMHash pairs
	CStringArray arrUserName;
	CStringArray arrLMHash;
	if (sInputType == "-h")
	{
		if (sInput.GetLength() != 32)
		{
			printf("invalid lmhash\n");
			return 0;
		}
		sInput.MakeLower();
		arrUserName.Add("UNKNOWN");
		arrLMHash.Add(sInput);
	}
	else if (sInputType == "-f")
	{
		if (LoadPwdDumpFile(sInput, arrUserName, arrLMHash) == FALSE)
		{
			printf("failed to open %s\n", sInput);
			return 0;
		}
		if (arrUserName.GetSize() == 0)
		{
			printf("no lmhash found\n");
			return 0;
		}
	}
	else
	{
		Usage();
		return 0;
	}

	// Some preparation
	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_IDLE);

	// FileList
	CStringArray arrPathName;
	GetFileList(sWildcharPathName, arrPathName);
	if (arrPathName.GetSize() == 0)
	{
		printf("no rainbow table found\n");
		return 0;
	}

	// Container
	CHashContainer hc;
	int i;
	for (i = 0; i < arrLMHash.GetSize(); i++)
	{
		hc.AddHash(arrLMHash[i].Left(16));
		hc.AddHash(arrLMHash[i].Mid(16));
	}
	CRainbowChainWalkContainer rcwc;

	// Search rainbow table
	for (i = 0; i < arrPathName.GetSize() && hc.AnyHashLeft(); i++)
	{
		SearchRainbowTable(arrPathName[i], hc, rcwc);
		printf("\n");
	}

	// Show statistics
	int nPlainTextFound, nHashTotal;
	hc.GetStatistics(nPlainTextFound, nHashTotal);
	printf("statistics\n");
	printf("------------------------------------------\n");
	printf("plaintext found:          %d of %d (%.2f%%)\n", nPlainTextFound, nHashTotal, 100.0f * nPlainTextFound / nHashTotal);
	printf("total disk access time:   %.2f s\n", g_fTotalDiskAccessTime);
	printf("total cryptanalysis time: %.2f s\n", g_fTotalCryptanalysisTime);
	printf("total chain walk step:    %d\n",     g_nTotalChainWalkStep);
	printf("total false alarm:        %d\n",     g_nTotalFalseAlarm);
	printf("total false alarm step:   %d\n",     g_nTotalFalseAlarmStep);
	printf("\n");

	// Show result
	printf("username      password\n");
	printf("------------------------------------------\n");
	for (i = 0; i < arrUserName.GetSize(); i++)
		printf("%-13s %s%s\n", arrUserName[i],
							   hc.GetPlainOfHash(arrLMHash[i].Left(16)),
							   hc.GetPlainOfHash(arrLMHash[i].Mid(16)));

	return 0;
}
