// TradeoffContext.cpp: implementation of the CTradeoffContext class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "TradeoffContext.h"
#include "HashRoutine.h"

#include <openssl/rand.h>
#pragma comment(lib, "libeay32.lib")

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////

BYTE CTradeoffContext::m_PlainCharset[1024];
int CTradeoffContext::m_nPlainCharsetLength;
CString CTradeoffContext::m_sPlainCharset;
CString CTradeoffContext::m_sPlainCharsetContent;
unsigned __int64 CTradeoffContext::m_nPlainSpaceUpToX[8];
unsigned __int64 CTradeoffContext::m_nPlainSpaceTotal;
int CTradeoffContext::m_nRainbowTableIndex;
unsigned __int64 CTradeoffContext::m_nReduceOffset;

//////////////////////////////////////////////////////////////////////

CTradeoffContext::CTradeoffContext()
{
}

CTradeoffContext::~CTradeoffContext()
{
}

void CTradeoffContext::GetAllCharset(CMapStringToString& mapCharsetNameCharset)
{
	mapCharsetNameCharset.RemoveAll();

	CStdioFile file;
	if (file.Open("charset.txt", CFile::modeRead | CFile::typeText))
	{
		CString sLine;
		while (file.ReadString(sLine))
		{
			int nIndex = sLine.Find("=");
			if (nIndex != -1)
			{
				// Filter comment
				sLine.TrimLeft();
				if (sLine[0] == '#')
					continue;

				// sCharsetName
				CString sCharsetName = sLine.Left(nIndex);
				sCharsetName.TrimLeft();
				sCharsetName.TrimRight();
				sCharsetName.MakeLower();
				if (sCharsetName == "")
					continue;
				// sCharsetName charset check, it is required because we use this name in file title
				BOOL fCharsetNameCheckOK = TRUE;
				int i;
				for (i = 0; i < sCharsetName.GetLength(); i++)
				{
					if (   (sCharsetName[i] >= 'a' && sCharsetName[i] <= 'z')
						|| (sCharsetName[i] >= '0' && sCharsetName[i] <= '9')
						||  sCharsetName[i] == '-')
						;
					else
						fCharsetNameCheckOK = FALSE;
				}
				if (fCharsetNameCheckOK == FALSE)
				{
					printf("invalid charset name %s in charset configuration file\n", sCharsetName);
					printf("line: %s\n", sLine);
					continue;
				}

				// sCharset
				CString sCharset = sLine.Mid(nIndex + 1);
				sCharset.TrimLeft();
				sCharset.TrimRight();
				if (sCharset == "" || sCharset == "[]")
					continue;
				if (sCharset[0] != '[' || sCharset[sCharset.GetLength() - 1] != ']')
				{
					printf("invalid charset syntax in charset configuration file\n");
					printf("line: %s\n", sLine);
					continue;
				}
				sCharset = sCharset.Mid(1, sCharset.GetLength() - 2);

				// Add to mapCharsetNameCharset
				//printf("%s = [%s]\n", sCharsetName, sCharset);
				mapCharsetNameCharset[sCharsetName] = sCharset;
			}
		}
		file.Close();
	}
	else
		printf("can't open charset configuration file\n");
}

BOOL CTradeoffContext::SetPlainCharset(CString sPlainCharset)
{
	// m_PlainCharset, m_nPlainCharsetLength, m_sPlainCharset, m_sPlainCharsetContent
	CMapStringToString mapCharsetNameCharset;
	GetAllCharset(mapCharsetNameCharset);
	CString sCharset = mapCharsetNameCharset[sPlainCharset];
	if (sCharset != "")
	{
		memcpy(m_PlainCharset, (LPCSTR)sCharset, sCharset.GetLength());
		m_nPlainCharsetLength = sCharset.GetLength();
	}
	else
		return FALSE;
	m_sPlainCharset = sPlainCharset;
	m_sPlainCharsetContent = sCharset;

	// m_nPlainSpaceUpToX
	m_nPlainSpaceUpToX[0] = 0;
	unsigned __int64 nTemp = 1;
	int i;
	for (i = 1; i <= 7; i++)
	{
		nTemp *= m_nPlainCharsetLength;
		m_nPlainSpaceUpToX[i] = m_nPlainSpaceUpToX[i - 1] + nTemp;
	}

	// m_nPlainSpaceTotal
	m_nPlainSpaceTotal = m_nPlainSpaceUpToX[7];

	// Dump all
	//for (i = 0; i <= 7; i++)
	//	printf("plain space up to %d: %I64u\n", i, m_nPlainSpaceUpToX[i]);

	return TRUE;
}

CString CTradeoffContext::GetPlainCharset()
{
	return m_sPlainCharset;
}

CString CTradeoffContext::GetPlainCharsetContent()
{
	return m_sPlainCharsetContent;
}

BOOL CTradeoffContext::SetRainbowTableIndex(int nRainbowTableIndex)
{
	// m_nReduceOffset
	if (nRainbowTableIndex < 0)
		return FALSE;
	m_nRainbowTableIndex = nRainbowTableIndex;
	m_nReduceOffset = 65536 * nRainbowTableIndex;

	return TRUE;
}

int CTradeoffContext::GetRainbowTableIndex()
{
	return m_nRainbowTableIndex;
}

void CTradeoffContext::GenerateRandomIndex()
{
	RAND_bytes((BYTE*)&m_nIndex, 8);
	m_nIndex = m_nIndex % m_nPlainSpaceTotal;
}

void CTradeoffContext::SetIndex(unsigned __int64 nIndex)
{
	m_nIndex = nIndex;
}

void CTradeoffContext::SetHash(BYTE Hash[8])
{
	memcpy(m_Hash, Hash, 8);
}

void CTradeoffContext::IndexToPlain()
{
	memset(m_Plain, 0, 7);

	int nPlainLength;
	int i;
	for (i = 6; i >= 0; i--)
	{
		if (m_nIndex >= m_nPlainSpaceUpToX[i])
		{
			nPlainLength = i + 1;
			break;
		}
	}

	unsigned __int64 nIndexOfX = m_nIndex - m_nPlainSpaceUpToX[nPlainLength - 1];

	/*
	// Slow version
	for (i = nPlainLength - 1; i >= 0; i--)
	{
		m_Plain[i] = m_PlainCharset[nIndexOfX % m_nPlainCharsetLength];
		nIndexOfX /= m_nPlainCharsetLength;
	}
	*/

	// Fast version
	for (i = nPlainLength - 1; i >= 0; i--)
	{
		if (nIndexOfX < 0x100000000I64)
			break;
		m_Plain[i] = m_PlainCharset[nIndexOfX % m_nPlainCharsetLength];
		nIndexOfX /= m_nPlainCharsetLength;
	}
	DWORD dwIndexOfX = (DWORD)nIndexOfX;
	for (; i >= 0; i--)
	{
		/*
		m_Plain[i] = m_PlainCharset[dwIndexOfX % m_nPlainCharsetLength];
		dwIndexOfX /= m_nPlainCharsetLength;
		*/
		int nTemp;
		dwIndexOfX /= m_nPlainCharsetLength;
		__asm mov nTemp, edx
		m_Plain[i] = m_PlainCharset[nTemp];
	}
}

void CTradeoffContext::PlainToHash()
{
	HashLM(m_Plain, m_Hash);
}

void CTradeoffContext::HashToIndex(int nPosition)
{
	m_nIndex = (*(unsigned __int64*)m_Hash + m_nReduceOffset + nPosition) % m_nPlainSpaceTotal;
}

unsigned __int64 CTradeoffContext::GetPlainSpaceTotal()
{
	return m_nPlainSpaceTotal;
}

unsigned __int64 CTradeoffContext::GetIndex()
{
	return m_nIndex;
}

CString CTradeoffContext::GetPlain()
{
	CString sRet;
	int i;
	for (i = 0; i < 7; i++)
	{
		if (m_Plain[i] != 0)
		{
			CString sByte;
			sByte.Format("%c", m_Plain[i]);
			sRet += sByte;
		}
		else
			sRet += " ";
	}

	return sRet;
}

CString CTradeoffContext::GetHash()
{
	CString sRet;
	int i;
	for (i = 0; i < 8; i++)
	{
		CString sByte;
		sByte.Format("%02x", m_Hash[i]);
		sRet += sByte;
	}

	return sRet;
}

BOOL CTradeoffContext::CheckHash(BYTE Hash[8])
{
	if (memcmp(m_Hash, Hash, 8) == 0)
		return TRUE;
	else
		return FALSE;
}

void CTradeoffContext::DumpPlain(unsigned __int64 nIndexFrom, unsigned __int64 nIndexTo)
{
	for (m_nIndex = nIndexFrom; m_nIndex <= nIndexTo; m_nIndex++)
	{
		IndexToPlain();
		printf("#%I64u %s\n", m_nIndex, GetPlain());
	}
}
