// RainbowTableGenerate.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "RainbowTableGenerate.h"
#include "../RainbowPublic/TradeoffContext.h"
#include "../RainbowPublic/Public.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// The one and only application object

CWinApp theApp;
int _tmain(int argc, TCHAR* argv[], TCHAR* envp[])
{
	if (!AfxWinInit(::GetModuleHandle(NULL), NULL, ::GetCommandLine(), 0))
		return 0;

	if (argc != 6)
	{
		Logo();

		printf("usage: rtgen plain_charset rainbow_table_index \\\n");
		printf("             rainbow_chain_length rainbow_chain_count file_title_suffix\n");
		printf("\n");
		printf("plain_charset:        use any charset name in charset.txt here\n");
		printf("rainbow_table_index:  index of the rainbow table\n");
		printf("rainbow_chain_length: length of the rainbow chain\n");
		printf("rainbow_chain_count:  count of the rainbow chain to generate\n");
		printf("file_title_suffix:    the string appended to the file title,\n");
		printf("                      add your comment of the generated rainbow table here\n");
		printf("\n");
		printf("example: rtgen alpha 0 100 16 bla\n");
		printf("         rtgen alpha-numeric 0 100 16 bla\n");
		return 0;
	}

	// Parse command line
	CString sPlainCharset = argv[1];
	if (CTradeoffContext::SetPlainCharset(sPlainCharset) == FALSE)
	{
		printf("invalid plain_charset\n");
		return 0;
	}
	printf("charset = [%s]\n", CTradeoffContext::GetPlainCharsetContent());
	int nRainbowTableIndex  = atoi(argv[2]);
	if (CTradeoffContext::SetRainbowTableIndex(nRainbowTableIndex) == FALSE)
	{
		printf("invalid rainbow_table_index\n");
		return 0;
	}
	int nRainbowChainLength = atoi(argv[3]);
	int nRainbowChainCount  = atoi(argv[4]);
	CString sFileTitleSuffix = argv[5];

	// Some preparation
	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_IDLE);

	// Open file
	CString sFileName;
	sFileName.Format("lm_%s_%d_%dx%d_%s.rt", sPlainCharset, nRainbowTableIndex, nRainbowChainLength, nRainbowChainCount, sFileTitleSuffix);
	CFile file;
	if (file.Open(sFileName, CFile::modeCreate | CFile::modeNoTruncate | CFile::modeWrite) == 0)
	{
		printf("failed to create file\n");
		return 0;
	}

	// Fix file length
	if (file.GetLength() % 16 != 0 && file.GetLength() % 16 != 8)
	{
		printf("rainbow table boundary check fail\n");
		file.Close();
		return 0;
	}
	if (file.GetLength() % 16 == 8)
		file.SetLength(file.GetLength() - 8);

	// Check existing chains
	int nRainbowChainCountInside = file.GetLength() / 16;
	if (nRainbowChainCountInside == nRainbowChainCount)
	{
		printf("precomputation of this rainbow table already finished\n");
		file.Close();
		return 0;
	}
	if (nRainbowChainCountInside != 0)
		printf("continuing from interrupted precomputing ...\n");
	file.SeekToEnd();

	// Generate rainbow table
	CTradeoffContext tc;
	DWORD dwTick = GetTickCount();
	int i;
	for (i = nRainbowChainCountInside; i < nRainbowChainCount; i++)
	{
		tc.GenerateRandomIndex();
		unsigned __int64 nIndex = tc.GetIndex();
		file.Write(&nIndex, 8);

		int nPosition;
		for (nPosition = 0; nPosition < nRainbowChainLength - 1; nPosition++)
		{
			tc.IndexToPlain();
			tc.PlainToHash();
			tc.HashToIndex(nPosition);
		}

		nIndex = tc.GetIndex();
		file.Write(&nIndex, 8);

		if ((i + 1) % 100000 == 0 || i + 1 == nRainbowChainCount)
		{
			printf("%9d of %9d rainbow chains generated (%d m %d s)\n", i + 1,
																	    nRainbowChainCount,
																		(GetTickCount() - dwTick) / 1000 / 60,
																		(GetTickCount() - dwTick) / 1000 % 60);
			dwTick = GetTickCount();
		}
	}

	// Close file
	file.Close();

	return 0;
}
