// RainbowTableSort.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "RainbowTableSort.h"
#include "Sort.h"
#include "../RainbowPublic/Public.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////

void Usage()
{
	Logo();

	printf("usage: rtsort rainbow_table_filename [-distinct]\n");
}

CWinApp theApp;
int _tmain(int argc, TCHAR* argv[], TCHAR* envp[])
{
	if (!AfxWinInit(::GetModuleHandle(NULL), NULL, ::GetCommandLine(), 0))
		return 0;

	if (argc != 2 && argc != 3)
	{
		Usage();
		return 0;
	}
	CString sFileName = argv[1];
	BOOL fDistinct = FALSE;
	if (argc == 3)
	{
		if (strcmp(argv[2], "-distinct") == 0)
			fDistinct = TRUE;
		else
		{
			Usage();
			return 0;
		}
	}

	// Open file
	CFile file;
	if (file.Open(sFileName, CFile::modeReadWrite) == 0)
	{
		printf("failed to open file\n");
		return 0;
	}

	if (file.GetLength() % 16 != 0)
	{
		printf("invalid file\n");
		file.Close();
		return 0;
	}

	// Sort
	MEMORYSTATUS ms;
	GlobalMemoryStatus(&ms);
	if (ms.dwAvailPhys >= file.GetLength())
	{
		int nRainbowChainCount = file.GetLength() / 16;
		RainbowChain* pChain = (RainbowChain*)new BYTE[file.GetLength()];
		if (pChain != NULL)
		{
			// Load file
			printf("loading rainbow table ...\n");
			file.Read(pChain, file.GetLength());

			// Sort file
			printf("sorting rainbow table ...\n");
			QuickSort(pChain, 0, nRainbowChainCount - 1);

			// Write file
			if (fDistinct == TRUE)
			{
				printf("writing sorted/distincted rainbow table ...\n");
				file.SetLength(0);
				unsigned __int64 nLastIndexE = 0xffffffffffffffff;
				int i;
				for (i = 0; i < nRainbowChainCount; i++)
				{
					if (pChain[i].nIndexE != nLastIndexE)
					{
						file.Write(&pChain[i], sizeof(RainbowChain));
						nLastIndexE = pChain[i].nIndexE;
					}
				}
			}
			else
			{
				printf("writing sorted rainbow table ...\n");
				file.SeekToBegin();
				file.Write(pChain, file.GetLength());
			}

			delete[] pChain;
		}
		else
			printf("memory allocation fail\n");
	}
	else
	{
		CString sTempFilePathName = sFileName + ".tmp";
		CFile tempfile;
		if (tempfile.Open(sTempFilePathName, CFile::modeCreate | CFile::modeReadWrite))
		{
			tempfile.SetLength(file.GetLength());
			ExternalSort(file, tempfile, fDistinct);
			tempfile.Close();
			printf("deleting temporary file ...\n");
			DeleteFile(sTempFilePathName);
		}
		else
			printf("failed to create temporary file, %u bytes free disk space required\n", file.GetLength());
	}

	// Close file
	file.Close();

	return 0;
}
