
#include "stdafx.h"
#include "Sort.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////

int QuickSortPartition(RainbowChain* pChain, int nLow, int nHigh)
{
	int nRandomIndex = nLow + (rand() * (RAND_MAX + 1) + rand()) % (nHigh - nLow + 1);
	RainbowChain TempChain;
	TempChain = pChain[nLow];
	pChain[nLow] = pChain[nRandomIndex];
	pChain[nRandomIndex] = TempChain;

	TempChain = pChain[nLow];
	unsigned __int64 nPivotKey = pChain[nLow].nIndexE;
	while (nLow < nHigh)
	{
		while (nLow < nHigh && pChain[nHigh].nIndexE >= nPivotKey)
			nHigh--;
		pChain[nLow] = pChain[nHigh];
		while (nLow < nHigh && pChain[nLow].nIndexE <= nPivotKey)
			nLow++;
		pChain[nHigh] = pChain[nLow];
	}
	pChain[nLow] = TempChain;
	return nLow;
}

void QuickSort(RainbowChain* pChain, int nLow, int nHigh)
{
	if (nLow < nHigh)
	{
		int nPivotLoc = QuickSortPartition(pChain, nLow, nHigh);
		QuickSort(pChain, nLow, nPivotLoc - 1);
		QuickSort(pChain, nPivotLoc + 1, nHigh);
	}
}

//////////////////////////////////////////////////////////////////////

BOOL PrepareSortedSegment(CArray<SortedSegment, SortedSegment>& arrSS, CFile& file, CFile& tempfile)
{
	// 4096M/x * 16K + xСֵ: x=8M,СֵΪ16M
	MEMORYSTATUS ms;
	GlobalMemoryStatus(&ms);
	if (ms.dwAvailPhys < 16 * 1024 * 1024)
	{
		printf("not enough free physical memory\n");
		return FALSE;
	}
	DWORD dwMemorySize = ms.dwAvailPhys / 16 * 16 - 8 * 1024 * 1024;

	// Allocate memory
	void* pMem = new BYTE[dwMemorySize];
	if (pMem == NULL)
	{
		printf("memory allocation fail\n");
		return FALSE;
	}

	// Prepare
	file.SeekToBegin();
	tempfile.SeekToBegin();
	int nSegmentIndex = 0;
	while (TRUE)
	{
		if (file.GetPosition() == file.GetLength())
			break;

		nSegmentIndex++;

		printf("reading segment #%d ...\n", nSegmentIndex);
		DWORD dwReadSize = file.Read(pMem, dwMemorySize);

		printf("sorting segment #%d ...\n", nSegmentIndex);
		QuickSort((RainbowChain*)pMem, 0, dwReadSize / sizeof(RainbowChain) - 1);

		SortedSegment ss;
		ss.nChainCount = 0;
		ss.nNextChainIndex = 0;
		ss.dwFileNextChainPos = tempfile.GetPosition();
		ss.nFileChainLeft = dwReadSize / sizeof(RainbowChain);
		arrSS.Add(ss);

		printf("writing sorted segment #%d to temporary file ...\n", nSegmentIndex);
		tempfile.Write(pMem, dwReadSize);
	}

	// Free memory
	delete[] pMem;

	return TRUE;
}

void LoadChainFromFile(CFile& tempfile, SortedSegment& ss)
{
	int nChainToRead;
	if (ss.nFileChainLeft < sizeof(ss.chain) / sizeof(RainbowChain))
		nChainToRead = ss.nFileChainLeft;
	else
		nChainToRead = sizeof(ss.chain) / sizeof(RainbowChain);

	tempfile.Seek(ss.dwFileNextChainPos, CFile::begin);
	tempfile.Read(ss.chain, nChainToRead * sizeof(RainbowChain));
	ss.nChainCount = nChainToRead;
	ss.nNextChainIndex = 0;
	ss.dwFileNextChainPos += nChainToRead * sizeof(RainbowChain);
	ss.nFileChainLeft -= nChainToRead;
}

void MergeSortedSegment(CArray<SortedSegment, SortedSegment>& arrSS, CFile& file, CFile& tempfile, BOOL fDistinct)
{
	if (fDistinct == TRUE)
		printf("writing sorted/distincted rainbow table ...\n");
	else
		printf("writing sorted rainbow table ...\n");

	// Merge and write back to the origin file
	file.SetLength(0);
	unsigned __int64 nLastIndexE = 0xffffffffffffffff;
	while (arrSS.GetSize() != 0)
	{
		if (arrSS[0].nChainCount == arrSS[0].nNextChainIndex)
			LoadChainFromFile(tempfile, arrSS[0]);
		int nMinIndex = 0;
		RainbowChain MinChain = arrSS[0].chain[arrSS[0].nNextChainIndex];

		int i;
		for (i = 1; i < arrSS.GetSize(); i++)
		{
			if (arrSS[i].nChainCount == arrSS[i].nNextChainIndex)
				LoadChainFromFile(tempfile, arrSS[i]);

			if (arrSS[i].chain[arrSS[i].nNextChainIndex].nIndexE < MinChain.nIndexE)
			{
				nMinIndex = i;
				MinChain = arrSS[i].chain[arrSS[i].nNextChainIndex];
			}
		}

		// Write chain when required
		if (fDistinct == TRUE)
		{
			if (MinChain.nIndexE != nLastIndexE)
			{
				file.Write(&MinChain, sizeof(RainbowChain));
				nLastIndexE = MinChain.nIndexE;
			}
		}
		else
			file.Write(&MinChain, sizeof(RainbowChain));

		// remove SortedSegment when required
		arrSS[nMinIndex].nNextChainIndex++;
		if (arrSS[nMinIndex].nNextChainIndex == arrSS[nMinIndex].nChainCount && arrSS[nMinIndex].nFileChainLeft == 0)
			arrSS.RemoveAt(nMinIndex);
	}
}

void ExternalSort(CFile& file, CFile& tempfile, BOOL fDistinct)
{
	CArray<SortedSegment, SortedSegment> arrSS;
	if (PrepareSortedSegment(arrSS, file, tempfile) == TRUE)
		MergeSortedSegment(arrSS, file, tempfile, fDistinct);
}
